!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: btreenode.gs,v 1.8 2008-01-09 19:02:30 stever Exp $
!
! Superclass Hierarchy:
!   BtreeNode, Array, SequenceableCollection, Collection, Object.
!
!=========================================================================

! class created in idxclasses.topaz

removeallmethods BtreeNode
removeallclassmethods BtreeNode
protectmethods

category: 'For Documentation Installation only'
classmethod: BtreeNode
installDocumentation

< unprotected >

| doc txt |
doc := GsClassDocumentation _newForPrivateGsClass: self.

txt := (GsDocText new) details:
'BtreeNode is the most general class for representing an node within a B-tree.'.
doc documentClassWith: txt.

txt := (GsDocText new) details:
'A SmallInteger that gives the number of elements contained in the node.'.
doc documentInstVar: #numElements with: txt.

self description: doc.
%

! ------------------- Class methods for BtreeNode
category: 'Instance Creation'
classmethod: BtreeNode
new

"Returns a new initialized instance with the correct size."
<protected>

^ (super new: self entrySize * self maxNumberOfElements) numElements: 0.
%

! ------------------- Instance methods for BtreeNode
category: 'Searching'
method: BtreeNode
_binarySearchCoveringKey: aKey totalOrder: aBoolean 

"Returns the index for the first entry in which aKey is found utilizing a
 binary search.  This is the first entry whose key >= aKey."
<protected>

| lowPt midPt highPt entrySize index |

entrySize := self entrySize.
lowPt := 1.
highPt := numElements.
[ lowPt <= highPt ] whileTrue: [
    midPt := (lowPt + highPt) quo: 2.
    index := midPt - 1 * entrySize + 2.

    (self _compareKey: aKey lessThanOrEqualToEntryAt: index)
        ifTrue: [ highPt := midPt - 1 ]
        ifFalse: [ lowPt := midPt + 1 ]
].

(self _compareKey: aKey lessThanOrEqualToEntryAt: index)
    ifFalse: [ ^ index + entrySize - 1 ].

^ index - 1
%

category: 'Searching'
method: BtreeNode
_binarySearchCoveringKey: aKey value: aValue 

"Returns the index for the entry in which aKey/aValue would be inserted.  This
 is the first entry which has its key > aKey.  If multiple entries have the
 same key, then aValue is used (since entries with duplicate keys are inserted
 by the value's OOP ordering).  If the receiver is empty, returns 1.  Uses a
 binary search."
<protected>

| lowPt midPt highPt entrySize index |

entrySize := self entrySize.
lowPt := 1.
highPt := numElements.
[ lowPt <= highPt ] whileTrue: [
    midPt := (lowPt + highPt) quo: 2.
    index := midPt - 1 * entrySize + 2.

    (self _compareKey: aKey value: aValue lessThanOrEqualToEntryAt: index)
        ifTrue: [ highPt := midPt - 1 ]
        ifFalse: [ lowPt := midPt + 1 ]
].

(self _compareKey: aKey value: aValue lessThanOrEqualToEntryAt: index)
    ifFalse: [ ^ index + entrySize - 1 ].

^ index -1
%

category: 'Searching'
method: BtreeNode
_binarySearchForKey: aKey value: aValue 

"Returns the 'value' index whose key equality matches aKey.  If no match is
 found, returns zero.  Uses a binary search."
<protected>

| key lowPt midPt highPt entrySize index |

numElements == 0 ifTrue: [ ^ 0 ].

entrySize := self entrySize.
lowPt := 1.
highPt := numElements.
[ lowPt <= highPt ] whileTrue: [
 midPt := (lowPt + highPt) quo: 2.
 index := midPt - 1 * entrySize + 2.
 key := self _at: index.

 (self _compareKey: aKey value: aValue equalToEntryAt: index)
  ifTrue: [ ^ index - 1 ].

 (self _compareKey: aKey value: aValue lessThanOrEqualToEntryAt: index)
  ifTrue: [ highPt := midPt - 1 ]
  ifFalse: [ lowPt := midPt + 1 ]
].

^ 0
%

category: 'Comparison Operators'
method: BtreeNode
_compareKey: aKey equalToEntryAt: index 

"Performs a = comparison between aKey and the entry whose key is at the given
 index.  The default implementation uses no encryption."
<protected>

^ aKey _idxCompareEqualTo: (self _at: index)
%

category: 'Comparison Operators'
method: BtreeNode
_compareKey: aKey lessThanEntryAt: index

"Perform a < comparison between aKey and the entry whose key is at the given
 index.  The default implementation uses no encryption."
<protected>

^ aKey _idxCompareLessThan: (self _at: index)
%

category: 'Comparison Operators'
method: BtreeNode
_compareKey: aKey lessThanOrEqualToEntryAt: index

"Performs a < comparison between aKey and the entry whose key is at the given
 index.  The default implementation uses no encryption."
<protected>

^ aKey _idxCompareLessThanOrEqual: (self _at: index)
%

category: 'Comparison Operators'
method: BtreeNode
_compareLastKeyLessThanOrEqualToKey: aKey value: aValue

"Returns whether the receivers last key/value pair is greater than or equal to
 the given key/value pair."
<protected>

numElements == 0 ifTrue: [ ^ false ].

" return aKey <= last key "
^ self
    _compareKey: aKey
    value: aValue
    lessThanOrEqualToEntryAt: self _lastKeyIndex
%

category: 'Updating'
method: BtreeNode
_createSibling

"Creates a new sibling node and moves half of the receiver's entries into the
 sibling.  Returns the new sibling."
<protected>

| newSibling sz splitIndex halfNumElements |
newSibling := self class new.
newSibling assignToSegment: self segment.
sz := numElements * self entrySize.

" splitIndex is where to split the receiver "
halfNumElements := numElements quo: 2.
splitIndex := halfNumElements * self entrySize + 1.

" copy the last half of entries into the new sibling "
newSibling _insertAt:1 from: self fromStart: splitIndex fromEnd: sz
           numToMoveDown: 0 .

" nil out last half of receiver "
self _deleteNoShrinkFrom: splitIndex to: sz.

" update the element counts in the receiver and the sibling "
newSibling numElements: (numElements - halfNumElements).
numElements := halfNumElements.

^ newSibling
%

category: 'Constants'
method: BtreeNode
_encryptionSize

"There is no default encryption."
<unprotected>

^ 0
%

category: 'Searching'
method: BtreeNode
_findCoveringIndexForKey: aKey totalOrder: bool1 

"Returns the index for the first entry in which aKey is found.  This is the
 first entry whose key >= aKey.  If the receiver's last key is greater than
 aKey, or if the receiver is empty, returns 0.  Uses a binary search."
<protected>

| index |
numElements == 0 ifTrue: [ ^ 0 ].

" faster if check results of binary search
(self _compareKey: aKey lessThanOrEqualToEntryAt: self _lastKeyIndex)
    ifFalse: [ ^ 0 ].
"

index := self _binarySearchCoveringKey: aKey totalOrder: bool1.
" check if not found "
^ index > self _lastKeyIndex
    ifTrue: [ 0 ]
    ifFalse: [ index ]
%

category: 'Updating'
method: BtreeNode
_insertKey: aKey value: aValue atIndex: insertionIndex

"Insert the key/value pair in the receiver.  The sender of this message must
 verify that the entry will fit in the receiver and provide the insertion
 index."
<protected>

| endIndex entrySize |

" see if any existing entries have to be moved "
entrySize := self entrySize .
endIndex := (numElements * entrySize) .

"move entries down to make room for a new entry, or add room at the end"
self _insertAt: insertionIndex 
             from: nil "from abstract Array of nils"
             fromStart: 1 
             fromEnd: entrySize 
             numToMoveDown: (endIndex - insertionIndex + 1) .


" add the new entry "
self _at: insertionIndex put: aValue.
self _at: (insertionIndex + 1) put: aKey.

self _insertEncryptionFor: aKey value: aValue startingAt: (insertionIndex + 2).

numElements := numElements + 1
%

category: 'Accessing'
method: BtreeNode
_lastIndex

"Returns the index of the last used slot in the node."
<protected>

^ numElements * self entrySize
%

category: 'Accessing'
method: BtreeNode
_lastKeyIndex

"Returns the index of the key of the last entry, or zero if the node is empty."
<protected>

^ numElements == 0
    ifTrue: [ 0 ]
    ifFalse: [ numElements - 1 * self entrySize + 2 ]
%

category: 'Constants'
method: BtreeNode
entrySize

"Returns the size of an entry with no encryption."
<unprotected>

^ 2
%

category: 'Constants'
classmethod: BtreeNode
entrySize

"Returns the size of an entry with no encryption."
<unprotected>

^ 2
%

category: 'Testing'
method: BtreeNode
isLeaf

"Returns whether the B-tree node is a leaf node.  The default return is false."
<unprotected>

^ false
%

category: 'Testing'
method: BtreeNode
isBtreeNode

"Returns true: the receiver is a B-tree node."
<unprotected>

^ true
%

category: 'Accessing'
method: BtreeNode
numElements

"Returns the value of the instance variable 'numElements'."
<protected>

^ numElements
%

category: 'Updating'
method: BtreeNode
numElements: newValue

"Modify the value of the instance variable 'numElements'."
<protected>

numElements := newValue
%

category: 'Updating'
method: BtreeNode
_incrementNumElements

"Increment the value of the instance variable 'numElements' by 1."
<protected>

numElements := numElements + 1
%

category: 'Constants'
method: BtreeNode
parentNodeClass

"Returns the class of node to be created as the parent when a split occurs."
<protected>

^ BtreeInteriorNode
%

category: 'Printing'
method: BtreeNode
prettyPrint

""
<protected>

| aStream |
aStream := PrintStream printingOn: String new.
self printOn: aStream withIndentationLevel: 0 includingValues: false.
^ aStream contents
%

category: 'Testing'
method: BtreeNode
_canCompare: aKey withClass: aClass

"Returns whether the receiver can make comparisons with the given key.
 Returns true since the comparison will actually be attempted."
<unprotected>

^ true
%

category: 'Accessing'
method: BtreeNode
at: anIndex

"Protected version of Object | at:"

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [^ self _errorNotIndexable].
(anIndex _isSmallInteger)
  ifTrue: [^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [^ self _errorNonIntegerIndex: anIndex].
self _primitiveFailed: #at: .
self _uncontinuableError
%

category: 'Accessing'
method: BtreeNode
instVarAt: anIndex

"Protected version of Object | instVarAt:"

<protected primitive: 811"with rcvrClass">

| numInstVars |
(anIndex _isInteger)
ifTrue:
[
   numInstVars:= self class allInstVarNames size.
   (numInstVars == 0)
   ifTrue:
      [ ^ self _error: #rtErrNoInstVars].
   ((anIndex >= 1) & (anIndex <= numInstVars))
   ifFalse: "out of bounds"
      [ ^ self _errorIndexOutOfRange: anIndex]
]
ifFalse: "anIndex not an integer"
   [ ^ self _errorNonIntegerIndex: anIndex
].

self _primitiveFailed: #instVarAt: .
self _uncontinuableError
%

category: 'Accessing'
method: BtreeNode
_at: anIndex

"Protected version of Object | _at:"

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
(anIndex _isSmallInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_at: .
self _uncontinuableError
%

category: 'Accessing'
method: BtreeNode
_basicAt: anIndex

"Protected version of Object | _basicAt:"

<protected primitive: 32>

(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
(anIndex _isSmallInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_basicAt: .
self _uncontinuableError
%

category: 'Accessing'
method: BtreeNode
_primitiveAt: anIndex

"Protected version of Object | _primitiveAt:"

<protected primitive: 89>

(anIndex _isSmallInteger)
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #_primitiveAt: .
self _uncontinuableError
%

category: 'Updating'
method: BtreeNode
at: anIndex put: aValue

"Protected version of Object | at:put:"

<protected primitive: 268>

(anIndex _isSmallInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue: [ ((aValue class ~~ SmallInteger) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #at:put: .
self _uncontinuableError
%

category: 'Updating'
method: BtreeNode
instVarAt: anIndex put: aValue

"Protected version of Object | instVarAt:put:"

<protected primitive: 804"with rcvrClass">
(self class allInstVarNames size == 0)
  ifTrue: [ ^ self _error: #rtErrNoInstVars].
(anIndex _isInteger)
  ifTrue: [((anIndex < 1) | (anIndex > self class allInstVarNames size))
        ifTrue: [ ^ self _errorIndexOutOfRange: anIndex]] "out of bounds"
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].

self _primitiveFailed: #instVarAt:put: .
self _uncontinuableError
%

category: 'Updating'
method: BtreeNode
size: anInteger

"Protected version of Object | size:"

<protected primitive: 803"with rcvrClass">

(self class isIndexable not) "not an indexable object"
  ifTrue: [ ^ self _errorNotIndexable].
(anInteger _isInteger)
   ifTrue: [ ^ self _errorIndexOutOfRange: anInteger]
   ifFalse: [ ^ self _errorNonIntegerIndex: anInteger].
self _primitiveFailed: #size: .
self _uncontinuableError
%

category: 'Updating'
method: BtreeNode
_basicAt: anIndex put: aValue

"Protected version of Object | _basicAt:put:"

<protected primitive: 268>

(anIndex _isSmallInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue class ~~ SmallInteger) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_basicAt:put: .
self _uncontinuableError
%

category: 'Updating'
method: BtreeNode
_at: anIndex put: aValue

"Protected version of Object | _at:put:"

<protected primitive: 268>

(anIndex _isSmallInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
(self class isIndexable) "not an indexable object"
  ifFalse: [ ^ self _errorNotIndexable].
((anIndex < 1) | (anIndex > (self size + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue class ~~ SmallInteger) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_at:put: .
self _uncontinuableError
%

category: 'Updating'
method: BtreeNode
_basicSize: anInteger

"Protected version of Object | _basicSize:"

<protected primitive: 803"with rcvrClass">

(self class isIndexable not) "not an indexable object"
  ifTrue: [ ^ self _errorNotIndexable].
(anInteger _isInteger)
   ifTrue: [ ^ self _errorIndexOutOfRange: anInteger]
   ifFalse: [ ^ self _errorNonIntegerIndex: anInteger].
self _primitiveFailed: #size: .
self _uncontinuableError
%

category: 'Updating'
method: BtreeNode
_primitiveAt: anIndex put: aValue

"Protected version of Object | _primitiveAt:put:"

<protected primitive: 281>

(anIndex _isSmallInteger)
  ifFalse: [ ^ self _errorNonIntegerIndex: anIndex].
((anIndex < 1) | (anIndex > (self _primitiveSize + 1))) "out of bounds"
  ifTrue: [ ^ self _errorIndexOutOfRange: anIndex].
(self class isBytes)
ifTrue:
  [ ((aValue class ~~ SmallInteger) _or: [ (aValue < 0) | (aValue > 255) ])
    ifTrue: [^ aValue _error: #rtErrExpectedByteValue].
  ].

self _primitiveFailed: #_primitiveAt:put: .
self _uncontinuableError
%

category: 'Reduced Conflict Support'
method: BtreeNode
_rcAt: anIndex

"Protected version of Object | _rcAt:"
<protected>

self _primitiveFailed: #_rcAt: .
self _uncontinuableError
%

category: 'Accessing'
method: BtreeNode
basicAt: anIndex

"Protected version of Object | basicAt:"
<protected>

^self _basicAt: anIndex
%

category: 'Updating'
method: BtreeNode
basicAt: anIndex put: aValue

"Protected version of Object | basicAt:put:"
<protected>

^self _basicAt: anIndex put: aValue
%

category: 'Copying'
method: BtreeNode
copyFrom: index1 to: index2 into: aSeqColl startingAt: destIndex

"Protected version of Array | copyFrom:to:into:startingAt:"

<protected primitive: 808"with rcvrClass">

^ super copyFrom: index1 to: index2 into: aSeqColl startingAt: destIndex
%

category: 'Copying'
method: BtreeNode
removeFrom: startIndex to: stopIndex

"Protected version of Array | removeFrom:to:"

<protected primitive: 801>

^ super removeFrom: startIndex to: stopIndex
%

category: 'Copying'
method: BtreeNode
insertAll: aSeqColl at: anIndex

"Protected version of Array | insertAll:at:"

<protected primitive: 806"with rcvrClass">

^ super insertAll: aSeqColl at: anIndex
%

category: 'Adding'
method: BtreeNode
add: newObject

"Protected version of Array | add:"

<protected primitive: 203>
self _primitiveFailed: #add: .
self _uncontinuableError
%

category: 'Adding'
method: BtreeNode
addLast: newObject

"Protected version of Array | addLast:"

<protected primitive: 203>
self _primitiveFailed: #addLast: .
self _uncontinuableError
%


category: 'Removing'
method: BtreeNode
_deleteNoShrinkFrom: startIndex to: stopIndex

"Protected version of Array | _deleteNoShrinkFrom:to: "

<protected primitive: 802 "with rcvrClass">
^ super _deleteNoShrinkFrom: startIndex to: stopIndex
%

category: 'Copying'
method: BtreeNode 
_insertAt: destOffset
from: anArray
fromStart: startOffset
fromEnd: endOffset
numToMoveDown: numToMove

"Protected version of Array | #_insertAt:from:fromStart:fromEnd:numToMoveDown:"

<protected primitive: 805"with rcvrClass">
^ super _insertAt: destOffset from: anArray fromStart: startOffset
            fromEnd: endOffset numToMoveDown: numToMove
%

category: 'Adding'
method: BtreeNode
_addAllFromNsc: aBag

"Protected version of Array | _addAllFromNsc:"

<protected primitive: 202>
self _primitiveFailed: #_addAllFromNsc: .
self _uncontinuableError
%

category: 'Indexing Support'
method: BtreeNode
maxNumWhenMerging

"Returns the maximum number of entries to add when building interior nodes
 during a merge sort on index creation."
<protected>

" ^ (self mergeThreshold + self class maxNumberOfElements) // 2 "
^ self class maxNumberOfElements - 5
%

category: 'Testing'
method: BtreeNode
_basicCanCompare: aKey withClass: aClass

"Returns true if the receiver can make comparisons with the given key, and
 false otherwise.  Used for basic nodes."
<protected>

(aKey isKindOfClass: aClass)
    ifTrue: [ ^ true ].

( (aKey isKindOf: Number) _and:
[ ( aClass inheritsFrom: Number) _or: [ aClass == Number ] ] )
    ifTrue: [ ^ true ].

( (aKey isKindOf: CharacterCollection) _and:
[ ( aClass inheritsFrom: CharacterCollection) _or:
[ aClass == CharacterCollection ] ] )
    ifTrue: [ ^ true ].

( (aKey isKindOf: AbstractCharacter) _and:
[ ( aClass inheritsFrom: AbstractCharacter) _or:
[ aClass == AbstractCharacter ] ] )
    ifTrue: [ ^ true ].

^ false
%
category: 'Constants'
method: BtreeNode
_emptyEntryArray

"Returns a two-element Array of nils."
<unprotected>

^ #( nil nil )
%

category: 'Formatting'
method: BtreeNode
printOn: aStream

"Puts a displayable representation of the receiver on the given stream."

"Default printing for non-protected access."

< unprotected >

aStream nextPutAll: self asString
%

! Deleted:   _obj: anObject isIdenticalTo: anotherObject


category: 'Constants'
classmethod: BtreeNode
maxNumberOfElements

"Returns the number of entries that are allowed in an interior node."
<protected>

^ 1000
%

category: 'Private'
method: BtreeNode
_canonicalizeSymbolAt: offset oldSymbol: oldSym newSymbol: newSym

"Override the method, but do nothing here because other updates will fix it."

<protected>

^self
%

unprotectmethods

