!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: canonstringdict.gs,v 1.7 2008-01-09 19:02:30 stever Exp $
!
! Superclass Hierarchy:
!   CanonicalStringDictionary, StringKeyValueDictionary, 
!   KeyValueDictionary, AbstractDictionary, Collection, Object.
!
!=========================================================================

removeallmethods CanonicalStringDictionary
removeallclassmethods CanonicalStringDictionary

category: 'For Documentation Installation only'
classmethod: CanonicalStringDictionary
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'A CanonicalStringDictionary is a StringKeyValueDictionary that provides
 protocol that is similar to Set in addition to its dictionary protocol.' .
doc documentClassWith: txt.

self description: doc.
%

category: 'Hashing'
method: CanonicalStringDictionary
hashFunction: aKey

"The hash function performs an operation on the value of the
 key (aKey) and returns a value in the range 1..tableSize."

^(aKey hash \\  self tableSize) + 1
%

category: 'Private'
method: CanonicalStringDictionary
compareKey: key1 with: key2

"Returns whether key1 is equivalent to key2."

 key1 == key2 ifTrue:[ ^ true ].

 (key1 _isSymbol ~~ key2 _isSymbol) 
    ifTrue:[ ^ key1 asString = key2 asString ]
   ifFalse:[ ^ key1 = key2 ].
%

category: 'Updating'
method: CanonicalStringDictionary
addAssociation: anAssociation

"Add the argument anAssociation to the receiver."

^ super add: anAssociation
%

category: 'Updating'
method: CanonicalStringDictionary
addAll: aCollection

"Adds elements of aCollection to the receiver. Returns aCollection."

aCollection == self ifTrue:[ ^ aCollection ].
(aCollection isKindOf: AbstractDictionary) 
  ifTrue:[ ^ super addAll: aCollection].
aCollection do:[ :aString | self add: aString ].
^ aCollection
%

category: 'Updating'
method: CanonicalStringDictionary
remove: aString

"Removes aString if present in the receiver and returns the removed value.  If
 aString is not present, generates an error."

^ self removeKey: aString
%

category: 'Updating'
method: CanonicalStringDictionary
remove: aString ifAbsent: aBlock

"Removes aString if present in the receiver and returns the removed value.  If
 aString is not present, returns the result of evaluating the zero
 argument Block aBlock."

^ self removeKey: aString ifAbsent: aBlock
%

category: 'Updating'
method: CanonicalStringDictionary
add: aString

"Adds aString if it is not already present in the receiver, and returns
 either aString or the canonical string already present."

| aValue |
(aString isKindOf: String) ifTrue:[
  aValue := self at: aString otherwise: nil .
  aValue == nil ifTrue:[
    aString immediateInvariant .
    self at: aString put: aString .
    ^ aString
    ].
  ^ aValue
  ].
"assume aString is an Association."
^ super add: aString  
%

category: 'Accessing'
method: CanonicalStringDictionary
includes: aString

"Returns true if the receiver contains aString as a key, false otherwise."

| aValue |
aValue := self at: aString otherwise: nil .
^ aValue ~~ nil
%

category: 'Accessing'
method: CanonicalStringDictionary
includesValue: aString

"Returns true if the receiver contains aString as a key, false otherwise."

| aValue |
aValue := self at: aString otherwise: nil .
^ aValue ~~ nil
%

category: 'Accessing'
method: CanonicalStringDictionary
_validateGroupString: aString

| aValue |
aValue := self at: aString otherwise: nil .
aValue == nil ifTrue:[ 
  aString _error: #segErrBadGroup .
  ].
^ aValue
%

category: 'Repository Conversion'
method: CanonicalStringDictionary
_migrateSymbolSet: aSymbolSet

"Using the receiver as the canonicalization dictionary, migrate aSymbolSet
 to an IdentitySet of canonical Strings, and return the IdentitySet."

| newSet |

((aSymbolSet isKindOf: SymbolSet) _or: [aSymbolSet isKindOf: ObsoleteSymbolSet])
  ifTrue:[
  newSet := IdentitySet new .
  aSymbolSet do:[ :aSym | | aStr groupStr |
    aStr := String withAll: aSym .
    groupStr := self add: aStr .
    newSet add: groupStr .
    ].
  newSet assignToSegment: aSymbolSet segment .
  ^ newSet 
  ].

^ aSymbolSet
%

category: 'Private'
method: CanonicalStringDictionary
_tableSize: newSize

"Sets the table size to a new value and creates all possible 
 collision buckets.  Used for rebuilding AllSymbols such that 
 the list of objects that make up the logical collection will
 seldom change."

| physSize cbCls mySeg |

(newSize <= 0)
  ifTrue: [newSize _error: #rtErrArgNotPositive  .  ^ self ].

"dereference old collision buckets from their parent to avoid problems
 when resolving symbol canonicalization at commit."
1 to: self _basicSize by: 2 do:[:j |
  | aBucket |
  (self basicAt: j) == nil ifTrue:[ "no key present, so check for a bucket"
    (aBucket := self basicAt: j + 1) ~~ nil ifTrue:[
       aBucket _removeAll . "dereference from parent"
      ].
    ].
  ].

physSize := newSize * 2 .
cbCls := self collisionBucketClass .
mySeg := self segment.

self nilOutAll.
self _basicSize: physSize .
"Create all collision buckets now."

1 to: physSize by: 2 do:[:i| 
 | newBucket |
 self _basicAt: i put: nil.
 newBucket := cbCls new.
 newBucket assignToSegment: mySeg ; keyValueDictionary: self.
 self _basicAt: (i + 1) put: newBucket.
].

numElements := 0.
numCollisions := 0.
collisionLimit := newSize // 4.
tableSize := newSize .
%

Category: 'Private'
method: CanonicalStringDictionary
_keysAsArray

"Collect return an Array containing all the keys of the receiver."

| result idx |
result := Array new: numElements .
idx := 1 .
1 to: tableSize * 2 by: 2 do: [ :tableIndex |
  | aKey collisionBkt |
  nil == (aKey := self _at: tableIndex) ifTrue: [
    nil == (collisionBkt := self _at: (tableIndex + 1)) ifFalse: [
      1 to: collisionBkt _basicSize by: 2 do: [ :j |
        nil == (aKey := collisionBkt _at: j) ifFalse: [
          result at: idx put: aKey .
          idx := idx + 1.
          ].
        ].
      ].
    ]
  ifFalse: [ result at: idx put: aKey .
             idx := idx + 1 ].
  ].
^ result
%

category: 'Private'
method: CanonicalStringDictionary
_saveKeysToHiddenSet

"Store all the keys of the receiver in hidden set 41.  
 This method is used as an optimization when rebuilding
 AllSymbols."

| idx buffer sys |
sys := System .
sys _hiddenSetReinit: 41 .
buffer := Array new: 2024.
idx := 1 .
1 to: tableSize * 2 by: 2 do: [ :tableIndex |
  | aKey collisionBkt |
  nil == (aKey := self _at: tableIndex) ifTrue: [
    nil == (collisionBkt := self _at: (tableIndex + 1)) ifFalse: [
      1 to: collisionBkt _basicSize by: 2 do: [ :j |
        nil == (aKey := collisionBkt _at: j) ifFalse: [
          buffer at: idx put: aKey .
          idx := idx + 1.
          idx == 2025 "buffer is full, dump to hidden set"
            ifTrue:[ sys _addAll: buffer to: 41.
                     idx := 1
              ].
          ].
        ].
      ].
    ]
  ifFalse: [ buffer at: idx put: aKey .
             idx := idx + 1.
             idx == 2025 "buffer is full, dump to hidden set"
               ifTrue:[ sys _addAll: buffer to: 41.
                       idx := 1 
               ].
  ].
].

"Add any remainder to hidden set"
(idx > 1)
  ifTrue:[ buffer size: (idx - 1) .
           sys _addAll: buffer to: 41].
buffer size: 0 .
^ self
%


category: 'Private'
method: CanonicalStringDictionary
_addKeysFromHiddenSet

"Destructively enumerate hidden set 41 and add them to the 
 receiver which should be AllSymbols."

| done sys |
done := false .
sys := System .
[ done ] whileFalse:[ | list |
  list := sys _hiddenSetEnumerate: 41 limit: 2024 .
  1 to: list size do:[:n| | sym |
    sym := list at: n.
    self at: sym put: sym .
  ].
  done := list size == 0.
  list size: 0.
].
^ self
%
 
category: 'Private'
method: CanonicalStringDictionary
_rebuildTable: newSize createAllBuckets: shouldCreateAllBuckets

"Rebuilds the hash table by saving the current state, initializing and
 changing the size of the table, and adding the key value pairs saved
 back to the hash dictionary.

 If shouldCreateAllBuckets is true, then all possible collision buckets
 are allocated when the table is rebuilt.

 *** This method is intended for use only with AllSymbols. ***

 Use System>>rebuildAllSymbols:createAllBuckets: to rebuild AllSymbols."

| saveCollLimit |

collisionLimit == 536870911 ifTrue:[
  "rehash disabled during symbol canonicalization."
  ^ self 
  ].

(self == (Globals at: #AllSymbols))
  ifFalse:[ nil error: 'illegal to send _rebuildTable: message to any object except AllSymbols'].

"We can take advantage of the fact that the keys and values
 in AllSymbols are identical.  So we only need to collect the keys."

self _saveKeysToHiddenSet . "Save to hidden set to avoid large object creation"
shouldCreateAllBuckets "Now re-size the table"
  ifTrue:[ self _tableSize: newSize ]
  ifFalse:[ self tableSize: newSize ].
saveCollLimit := collisionLimit .
collisionLimit := 536870911 . "prevent recursive rebuild"
self _addKeysFromHiddenSet .
collisionLimit := saveCollLimit .
^self
%
